#!/usr/bin/env python

from distutils.core import setup

setup(name='PwPEG',
      version='0.3.1',
      description='Python PEG Parsing Framework',
      author='Christophe Eymard',
      author_email='christophe.eymard@ravelsoft.com',
      url='http://pwpeg.github.com/',
      packages=['pwpeg'],
      scripts=['scripts/pwpeg']
     )
